use serde::{
    de::{self, Deserializer},
    Deserialize, Serialize,
};
use std::convert::TryFrom;
use thiserror::Error;
use types::{Event, SubscriptionId};

pub mod client;
pub mod types;

#[derive(Debug, Error)]
pub enum Error {}

#[derive(Serialize, Debug, Clone)]
pub enum Message {
    Request(client::Request),
    EventFromClient(types::Event),
    EventFromServer(types::SubscriptionId, types::Event),
    Close(types::SubscriptionId),
    Notice(String),
}

impl Message {
    pub fn as_request(&self) -> Option<&client::Request> {
        match self {
            Self::Request(x) => Some(x),
            _ => None,
        }
    }

    pub fn as_event_from_server(&self) -> Option<(&types::SubscriptionId, &types::Event)> {
        match self {
            Self::EventFromServer(id, event) => Some((id, event)),
            _ => None,
        }
    }

    pub fn as_event_from_client(&self) -> Option<&types::Event> {
        match self {
            Self::EventFromClient(event) => Some(event),
            _ => None,
        }
    }

    pub fn as_notice(&self) -> Option<&str> {
        match self {
            Self::Notice(x) => Some(x),
            _ => None,
        }
    }

    pub fn as_close_subscription_id(&self) -> Option<&types::SubscriptionId> {
        match self {
            Self::Close(x) => Some(x),
            _ => None,
        }
    }
}

impl<'de> de::Deserialize<'de> for Message {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let array: Vec<serde_json::Value> = Deserialize::deserialize(deserializer)?;
        if array.is_empty() {
            return Err(de::Error::custom(
                "Invalid array length, expecting at least one",
            ));
        }

        let tag = array
            .get(0)
            .ok_or_else(|| de::Error::custom("Invalid array length, expecting at least one"))?
            .as_str()
            .ok_or_else(|| de::Error::custom("Invalid type for element 0 of the array"))?;

        match tag {
            "EVENT" => match array.len() {
                3 => {
                    let subscription_id: SubscriptionId = array[1]
                        .as_str()
                        .map(TryFrom::try_from)
                        .transpose()
                        .map_err(|e: types::subscription_id::Error| {
                            de::Error::custom(e.to_string())
                        })?
                        .ok_or_else(|| de::Error::custom("Invalid subscription ID"))?;

                    let event: Event = serde_json::from_value(array[2].clone())
                        .map_err(|e: serde_json::Error| de::Error::custom(e.to_string()))?;

                    event
                        .is_valid()
                        .map_err(|e| de::Error::custom(e.to_string()))?;

                    Ok(Self::EventFromServer(subscription_id, event))
                }
                2 => {
                    let event: Event = serde_json::from_value(array[1].clone())
                        .map_err(|e: serde_json::Error| de::Error::custom(e.to_string()))?;

                    event
                        .is_valid()
                        .map_err(|e| de::Error::custom(e.to_string()))?;

                    Ok(Self::EventFromClient(event))
                }
                _ => Err(de::Error::custom("Invalid length for EVENT")),
            },

            "NOTICE" => Ok(Self::Notice(
                serde_json::from_value(
                    array
                        .get(1)
                        .ok_or_else(|| de::Error::custom("Missing element 1 of the array"))?
                        .clone(),
                )
                .map_err(|e: serde_json::Error| de::Error::custom(e.to_string()))?,
            )),
            "CLOSE" => Ok(Self::Close(
                serde_json::from_value(
                    array
                        .get(1)
                        .ok_or_else(|| de::Error::custom("Missing element 1 of the array"))?
                        .clone(),
                )
                .map_err(|e: serde_json::Error| de::Error::custom(e.to_string()))?,
            )),
            "REQ" => {
                let subscription_id = array
                    .get(1)
                    .ok_or_else(|| de::Error::custom("Missing element 1 in the array"))?
                    .as_str()
                    .ok_or_else(|| {
                        de::Error::custom("Invalid type for element 1, expecting a string")
                    })?
                    .try_into()
                    .map_err(|e: types::subscription_id::Error| {
                        de::Error::custom(format!("Invalid subscription id: {}", e))
                    })?;

                Ok(Self::Request(client::Request {
                    subscription_id,
                    filters: if array.len() > 2 {
                        serde_json::from_value::<Vec<types::Filter>>(serde_json::Value::Array(
                            array[2..].to_owned(),
                        ))
                        .map_err(|e: serde_json::Error| de::Error::custom(e.to_string()))?
                    } else {
                        vec![]
                    },
                }))
            }
            tag => Err(de::Error::custom(format!("{} is not a support tag", tag))),
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn unsupported() {
        let json = "[\"CLOSEX\", \"foo\"]";
        let message: Result<Message, _> = serde_json::from_str(json);
        assert!(message.is_err());
    }

    #[test]
    fn event_from_server() {
        let json = "[\"EVENT\",\"640bddcc93eae\",{\"content\":\"🤙\",\"created_at\":1676637072,\"id\":\"a3eaa71e4f46c1a69ac0596ae7c2af35807fc0b0d3b208b79a36eef67ef51743\",\"kind\":7,\"pubkey\":\"b2815682cfc83fcd2c3add05785cf4573dd388457069974cc6d8cca06b3c3b78\",\"sig\":\"93a3e9c4f6cb9a704885c4f77f6d7b16153e8eedc967603602606147f8c78d426f547d54120b80b33bd2101de638c06f2932df4daf53d66ca9b1341f2fd45729\",\"tags\":[[\"p\",\"8fe53b37518e3dbe9bab26d912292001d8b882de9456b7b08b615f912dc8bf4a\",\"\",\"mention\"],[\"e\",\"eb278e983fcedbb0d143c4250c879d078d037586c5dca8e1cf1a104f9846a460\"],[\"p\",\"2bda4f03446bc1c6ff00594e350a1e7ec57fb9cdc4a7b52694223d56ce0599e9\"]]}]";
        let message: Message = serde_json::from_str(json).expect("valid message");
        assert!(message.as_event_from_server().is_some());

        let (_, event) = message.as_event_from_server().expect("event");
        assert_eq!("🤙".to_owned(), event.content.content);
    }

    #[test]
    fn close() {
        let json = "[\"CLOSE\", \"foo\"]";
        let message: Message = serde_json::from_str(json).expect("valid message");
        let subscription_id = message
            .as_close_subscription_id()
            .expect("valid subscription_id");

        assert_eq!("foo".to_owned(), subscription_id.to_string());
    }
}
